import os
import socket
import ssl
import threading
from pathlib import Path
from typing import List

import pytest
from testcontainers.core.network import Network

from pyrelukko import RelukkoClient
from pyrelukko.testcontainers import RelukkoContainer, RelukkoDbContainer

SCRIPT_DIR = Path(__file__).parent.absolute()


@pytest.fixture(scope="session")
def relukko_backend():
    if os.environ.get("CI_HAS_RELUKKO"):
        # Gitlab does not allow DinD networks, so Relukko runs as a service
        # in the background of the job.
        relukko = RelukkoClient(
            base_url=os.environ['CI_RELUKKO_BASE_URL'],
            api_key=os.environ['CI_RELUKKO_API_KEY']
        )
        yield relukko, None
    else:
        with Network() as rl_net:
            with RelukkoDbContainer(net=rl_net,
                image="postgres:16", hostname="relukkodb") as _db:
                db_url = "postgresql://relukko:relukko@relukkodb/relukko"
                with RelukkoContainer(rl_net, db_url=db_url) as backend:
                    relukko = RelukkoClient(
                        base_url=backend.get_api_url(), api_key="somekey")
                    yield relukko, backend

@pytest.fixture(scope="function")
def tls_listener():

    certfile = SCRIPT_DIR / "cert" / "relukko.crt"
    keyfile = SCRIPT_DIR / "cert" / "relukko.key"

    def run_server(port_info: List, keep_running: threading.Event):
        # Create a TCP socket
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0) as sock:
            sock.bind(("127.0.0.1", 0))
            sock.listen(5)  # Listen for incoming connections

            _, assigned_port = sock.getsockname()
            port_info.append(assigned_port)

            # Wrap the socket with TLS
            context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            context.load_cert_chain(certfile=certfile, keyfile=keyfile)
            with context.wrap_socket(sock, server_side=True) as ssock:

                ssock.settimeout(1)

                while keep_running.is_set():
                    try:
                        client_socket, _ = ssock.accept()
                        client_socket.sendall(b"Hello, TLS Client!")
                        client_socket.close()
                    except Exception as _:
                        continue

    port_info = []
    keep_running = threading.Event()
    keep_running.set()
    thread = threading.Thread(target=run_server, args=(port_info, keep_running))
    thread.start()

    while not port_info:
        pass

    yield thread, port_info[0]

    keep_running.clear()
    thread.join()



# pylint: skip-file
"""
Testcontainers to be used in PyTests, for a fixture see tests/conftest.py
(relukko_backend):

@pytest.fixture(scope="session")
def relukko_backend():
    with Network() as rl_net:
        with RelukkoDbContainer(net=rl_net,
            image="postgres:16", hostname="relukkodb") as _db:
            db_url = "postgresql://relukko:relukko@relukkodb/relukko"
            with RelukkoContainer(rl_net, db_url=db_url) as backend:
                relukko = RelukkoClient(
                    base_url=backend.get_api_url(), api_key="somekey")
                yield relukko, backend
"""
import socket

from testcontainers.core.network import Network
from testcontainers.core.waiting_utils import wait_container_is_ready
from testcontainers.generic import ServerContainer
from testcontainers.postgres import PostgresContainer


class RelukkoContainer(ServerContainer):
    def __init__(self, net: Network,
                 image="registry.gitlab.com/relukko/relukko:latest", db_url=None):
        self.db_url = db_url
        self.net = net
        super(RelukkoContainer, self).__init__(image=image, port=3000)

    def _configure(self):
        self.with_env("DATABASE_URL", self.db_url)
        self.with_env("RELUKKO_API_KEY", "somekey")
        self.with_env("RELUKKO_USER", "relukko")
        self.with_env("RELUKKO_PASSWORD", "relukko")
        self.with_env("RELUKKO_BIND_ADDR", "0.0.0.0")
        self.with_network(self.net)

    def get_api_url(self) -> str:
        return f"http://localhost:{self.get_exposed_port(3000)}"

    def _create_connection_url(self) -> str:
        return f"{self.get_api_url()}/healthchecker"


class RelukkoDbContainer(PostgresContainer):
    def __init__(
            self, net: Network,
            image: str = "postgres:latest",
            port: int = 5432,
            username: str | None = None,
            password: str | None = None,
            dbname: str | None = None,
            driver: str | None = "psycopg2",
            **kwargs) -> None:
        self.net = net
        super().__init__(image, port, username, password, dbname, driver, **kwargs)

    def _configure(self) -> None:
        self.with_env("POSTGRES_USER", "relukko")
        self.with_env("POSTGRES_PASSWORD", "relukko")
        self.with_env("POSTGRES_DB", "relukko")
        self.with_network(self.net)

    @wait_container_is_ready()
    def _connect(self) -> None:
        packet = bytes([
            0x00, 0x00, 0x00, 0x52, 0x00, 0x03, 0x00, 0x00, 
            0x75, 0x73, 0x65, 0x72, 0x00, 0x72, 0x65, 0x6c, 
            0x75, 0x6b, 0x6b, 0x6f, 0x00, 0x64, 0x61, 0x74, 
            0x61, 0x62, 0x61, 0x73, 0x65, 0x00, 0x72, 0x65, 
            0x6c, 0x75, 0x6b, 0x6b, 0x6f, 0x00, 0x61, 0x70, 
            0x70, 0x6c, 0x69, 0x63, 0x61, 0x74, 0x69, 0x6f, 
            0x6e, 0x5f, 0x6e, 0x61, 0x6d, 0x65, 0x00, 0x70, 
            0x73, 0x71, 0x6c, 0x00, 0x63, 0x6c, 0x69, 0x65, 
            0x6e, 0x74, 0x5f, 0x65, 0x6e, 0x63, 0x6f, 0x64, 
            0x69, 0x6e, 0x67, 0x00, 0x55, 0x54, 0x46, 0x38, 
            0x00, 0x00
        ])
        port = self.get_exposed_port(self.port)
        with socket.create_connection(("localhost", port)) as sock:
            sock.send(packet)
            buf = sock.recv(40)
            if len(buf) == 0 and "SCRAM-SHA" not in buf:
                raise ConnectionError



"""
TBD
"""
import asyncio
import json
import logging
import os
import ssl
import threading
import time
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Union

import requests
from urllib3.util import Url, parse_url
from urllib3.util.retry import Retry
from websockets import ConnectionClosed as WsConnectionClosed
from websockets.asyncio.client import connect as ws_connect

from .retry import retry

SSL_KWARGS = [
    'check_hostname',
    'hostname_checks_common_name',
    'verify_mode',
    'verify_flags',
    'options',
]

RETRY_KWARGS = [
    'tries',
    'delay',
    'backoff',
    'max_delay',
    'exceptions',
]

OWN_KWARGS = [
    'acquire_wait_for_timeout',
    'acquire_modulo',
    'disable_websocket',
    'raise_when_acquire_fails',
    'ws_ping_interval',
    'ws_ping_timeout',
    'ws_wait_for_timeout',
]
OWN_KWARGS.extend(RETRY_KWARGS)

logger = logging.getLogger(__name__)


class RelukkoDoRetry(Exception):
    """
    Exception thrown on some errors which we still want to retry
    """

# pylint: disable=too-many-instance-attributes
class RelukkoClient:
    """
    TBD
    """
    def __init__(
            self, base_url: Union[Url, str], api_key:str, **kwargs) -> None:
        """
        TBD
        """
        self.session = requests.Session()
        self.api_key = api_key
        self.tries=4
        self.delay=5
        self.backoff=2.0
        self.max_delay=None
        self.exceptions = (
            requests.ConnectionError,
            RelukkoDoRetry,
        )
        self._setup_session(api_key, **kwargs)
        self._setup_http_adapters_retry(**kwargs)
        self.acquire_modulo = 100
        self.acquire_wait_for_timeout = 2
        self.disable_websocket = False
        self.raise_when_acquire_fails = True
        self.ws_ping_interval = 60
        self.ws_ping_timeout = 20
        self.ws_wait_for_timeout = 2
        self._setup_pyrelukko_kwargs(**kwargs)

        self.base_url = self._setup_base_url(base_url)
        self.ws_url = self._setup_ws_url(str(self.base_url))
        self.ssl_ctx: ssl.SSLContext = None
        self._setup_ssl_ctx(**kwargs)

        # event for websocket thread to signal it got a message
        self.message_received = threading.Event()
        # As long as it's set the websocket thread runs
        self.ws_running = threading.Event()
        self.ws_listener: threading.Thread = None

    def reconfigure_relukko(
            self, base_url: Union[Url, str]=None, api_key: str=None, **kwargs):
        """
        TBD
        """
        self.api_key = api_key or self.api_key
        self._setup_session(self.api_key, **kwargs)
        self._setup_http_adapters_retry(**kwargs)
        self.base_url = self._setup_base_url(base_url or self.base_url)
        self.ws_url = self._setup_ws_url(str(self.base_url))
        self._setup_ssl_ctx(**kwargs)
        self._setup_pyrelukko_kwargs(**kwargs)

    def _setup_pyrelukko_kwargs(self, **kwargs):
        for kwarg in OWN_KWARGS:
            setattr(
                self,
                kwarg,
                kwargs.get(kwarg, getattr(self, kwarg))
            )

    def _setup_http_adapters_retry(self, **kwargs):
        for _, http_adapter in self.session.adapters.items():
            http_retry: Retry = http_adapter.max_retries
            for key, value in kwargs.items():
                if hasattr(http_retry, key):
                    setattr(http_retry, key, value)
            http_adapter.max_retries = http_retry

    def _setup_session(self, api_key: str, **kwargs):
        self.session.headers['X-api-Key'] = api_key
        for key, value in kwargs.items():
            if hasattr(self.session, key):
                setattr(self.session, key, value)

    def _setup_ssl_ctx(self, **kwargs) -> Union[ssl.SSLContext, None]:
        if self.ws_url.scheme == "wss":
            if self.ssl_ctx is None:
                self.ssl_ctx = ssl.create_default_context(
                    ssl.Purpose.SERVER_AUTH)
            for kwarg in SSL_KWARGS:
                setattr(
                    self.ssl_ctx,
                    kwarg,
                    kwargs.get(kwarg, getattr(self.ssl_ctx, kwarg)))

            # Try to behave like requests library and take *_CA_BUNDLE env vars
            # into account.
            ca_bundle = (
                os.environ.get("REQUESTS_CA_BUNDLE")
                or os.environ.get("CURL_CA_BUNDLE"))

            ca_bundle_file = None
            ca_bundle_path = None
            if ca_bundle is not None:
                ca_bundle = Path(ca_bundle)
                ca_bundle_file = ca_bundle if ca_bundle.is_file() else None
                ca_bundle_path = ca_bundle if ca_bundle.is_dir() else None

            # values from kwargs take precedence env vars
            ca_file = kwargs.get('cafile', ca_bundle_file)
            ca_path = kwargs.get('capath', f"{ca_bundle_path}/")
            ca_data = kwargs.get('cadata')

            if ca_file or ca_path or ca_data:
                self.ssl_ctx.load_verify_locations(
                    cafile=ca_file, capath=ca_path, cadata=ca_data)
        else:
            self.ssl_ctx = None

    def _setup_ws_url(self, ws_url: str) -> Url:
        url = ws_url.replace("http", "ws", 1)
        return parse_url(f"{url}/ws/broadcast")

    def _setup_base_url(self, base_url: Union[Url, str]) -> Url:
        if isinstance(base_url, str):
            base_url = parse_url(base_url)
        if not isinstance(base_url, Url):
            raise ValueError("must be URL or string!")

        return base_url

    async def _websocket_listener(self):
        """
        The WebSocket thread, which waits for messages from Relukko and
        notifies the HTTP thread in case deletions happend, so the HTTP
        can retry to get the lock. Does not verify the wanted lock got
        deleted yet.
        """
        additional_headers = { "X-API-KEY": self.api_key }
        async with ws_connect(
            str(self.ws_url),
            additional_headers=additional_headers,
            ssl=self.ssl_ctx,
            logger=logger,
            ping_interval=self.ws_ping_interval,
            ping_timeout=self.ws_ping_timeout,
        ) as websocket:
            while self.ws_running.is_set():
                try:
                    ws_message = await asyncio.wait_for(
                        websocket.recv(), timeout=self.ws_wait_for_timeout)
                    if ws_message:
                        logger.debug("Received message: '%s'", ws_message)
                        msg: Dict = json.loads(ws_message)
                        if msg.get('deleted'):
                            # Signal the HTTP thread to wake up
                            self.message_received.set()
                except TimeoutError:
                    # no messages, try in a moment again...
                    time.sleep(0.5)
                except WsConnectionClosed:
                    logger.error("Lost WS connection!")
                    continue

    def _acquire_relukko(
            self, url: Union[Url, str], max_run_time: int,
            payload: Dict, _thread_store: List):
        """
        The HTTP thread which tries to create the Relukko lock.
        """

        start_time = time.time()
        loop_counter = 0
        got_message = False
        res = None
        while True:
            elapsed_time = time.time() - start_time
            if elapsed_time > max_run_time:
                self.ws_running.clear()
                _thread_store.insert(0, None)
                break
            # If self.message_recieved is True try to get lock ASAP!
            # Otherwise only in every Xth run in case websocket broke.
            if got_message or loop_counter % self.acquire_modulo == 0:
                try:
                    res = self._make_request(
                        url=url, method="POST", payload=payload)
                except (
                    *self.exceptions, RuntimeError, requests.HTTPError) as e:
                    logger.warning("Last exception was: %s!\nGiving up!", e)
                    _thread_store.insert(0, e)
                    break
            loop_counter += 1
            if res is None:
                # Conflict 409
                got_message = self.message_received.wait(
                    timeout=self.acquire_wait_for_timeout)
                self.message_received.clear()
                continue

            _thread_store.insert(0, res)
            self.ws_running.clear()
            break

    def _check_response(self, response: requests.Response):
        match response.status_code:
            case 200 | 201 | 404 | 422:
                return response.json()
            case 400 | 403:
                err = response.json()
                logger.warning("4xx HTTP Error [%d](%s) - %s:%s",
                    response.status_code, response.reason,
                    str(err.get('status')), err.get('message'))
                response.raise_for_status()
            case 409:
                err = response.json()
                logger.warning("409 HTTP Error [%d](%s) - %s:%s",
                    response.status_code, response.reason,
                    str(err.get('status')), err.get('message'))
                return None
            case 500 | 502 | 503 | 504:
                logger.warning("[%d](%s) %s",
                    response.status_code, response.reason, response.text)
                raise RelukkoDoRetry(
                    f"5xx HTTP Error: [{response.status_code}]"
                    f"({response.reason})")
            case _:
                logger.warning("[%d](%s) %s",
                    response.status_code, response.reason, response.text)
                raise RuntimeError(
                    f"Give up: [{response.status_code}]({response.reason})")

    def _make_request(
            self,
            url: str,
            method: str,
            payload: Dict=None) -> requests.Response:


        @retry(logger, exceptions=self.exceptions, tries=self.tries,
               delay=self.delay, backoff=self.backoff,
               max_delay=self.max_delay)
        def _do_request():
            response = self.session.request(
                method=method,
                url=url,
                json=payload,
            )
            return self._check_response(response)

        return _do_request()

    def acquire_relukko(self, lock_name, creator, max_run_time):
        """
        TBD
        """
        payload = {
            "lock_name": lock_name,
            "creator": creator,
        }

        url = f"{self.base_url}/v1/locks/"

        if not self.disable_websocket:
            self.ws_running.set()
            self.ws_listener = threading.Thread(
                target=asyncio.run, args=(self._websocket_listener(),))
            self.ws_listener.start()

        thread_store = []
        http_thread = threading.Thread(
            target=self._acquire_relukko,
            args=(url, max_run_time, payload, thread_store))
        http_thread.start()
        http_thread.join()

        if not self.disable_websocket:
            self.ws_listener.join()

        if thread_store:
            if (
                self.raise_when_acquire_fails
                and isinstance(thread_store[0], Exception)
            ):
                raise thread_store[0]
            return thread_store[0]
        return None

    def get_lock(self, lock_id: str) -> Dict:
        """
        TBD
        """
        url = f"{self.base_url}/v1/locks/{lock_id}"
        return self._make_request(url, "GET")

    def get_locks(self)  -> List:
        """
        TBD
        """
        url = f"{self.base_url}/v1/locks/"
        return self._make_request(url, "GET")

    def update_relukko(
            self, lock_id: str, creator: str=None, expires_at: datetime=None):
        """
        TBD
        """
        if isinstance(expires_at, datetime):
            expires_at = expires_at.isoformat()
        elif expires_at is not None:
            raise ValueError("has to be datetime!")

        payload = {
            "creator": creator,
            "expires_at": expires_at,
        }
        url = f"{self.base_url}/v1/locks/{lock_id}"
        return self._make_request(url, "PUT", payload)

    def delete_relukko(self, lock_id: str):
        """
        TBD
        """
        url = f"{self.base_url}/v1/locks/{lock_id}"
        return self._make_request(url, "DELETE")

    def keep_relukko_alive(self, lock_id: str):
        """
        TBD
        """
        url = f"{self.base_url}/v1/locks/{lock_id}/keep_alive"
        return self._make_request(url, "GET")

    def keep_relukko_alive_put(self, lock_id: str, seconds: int):
        """
        TBD
        """
        url = f"{self.base_url}/v1/locks/{lock_id}/keep_alive"
        payload = {
            "seconds": seconds
        }
        return self._make_request(url, "PUT", payload)

    def add_to_expires_at_time(self, lock_id: str):
        """
        TBD
        """
        url = f"{self.base_url}/v1/locks/{lock_id}/add_to_expire_at"
        return self._make_request(url, "GET")

    def add_to_expires_at_time_put(self, lock_id: str, seconds: int):
        """
        TBD
        """
        url = f"{self.base_url}/v1/locks/{lock_id}/add_to_expire_at"
        payload = {
            "seconds": seconds
        }
        return self._make_request(url, "PUT", payload)



"""
TBD
"""
import time
from functools import wraps


# pylint: disable=too-many-arguments,too-many-positional-arguments
def retry(logger, exceptions, tries=4, delay=5,
          backoff=2.0, max_delay=None):
    """
    Retry calling the decorated function using an exponential backoff.
    https://www.calazan.com/retry-decorator-for-python-3/

    :param exceptions: The exception to check. may be a tuple of
     exceptions to check.
    :param tries: Number of times to try (not retry) before giving up.
    :param delay: Initial delay between retries in seconds.
    :param backoff: Backoff multiplier (e.g. value of 2 will double the delay
     each retry).
    :param max_delay: maximum value for delay
    """
    def deco_retry(f):

        @wraps(f)
        def f_retry(*args, **kwargs):
            remaining_tries, retry_delay = tries, delay
            while remaining_tries > 1:
                try:
                    return f(*args, **kwargs)
                except exceptions:
                    remaining_tries -= 1
                    logger.warning('(%i/%i): Retrying in %i seconds...',
                        tries - remaining_tries,
                        tries,
                        retry_delay,
                    )
                    time.sleep(retry_delay)
                    if max_delay is not None:
                        retry_delay = min(retry_delay*backoff, max_delay)
                    else:
                        retry_delay *= backoff
            return f(*args, **kwargs)
        return f_retry  # true decorator
    return deco_retry



import logging
import time
import ssl
import socket
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Dict

import pytest
import requests
from requests_mock import ANY
from requests.adapters import HTTPAdapter
from requests.cookies import cookiejar_from_dict
from urllib3.util import parse_url

from pyrelukko import RelukkoClient
from pyrelukko.pyrelukko import RelukkoDoRetry
from pyrelukko.retry import retry


SCRIPT_DIR = Path(__file__).parent.absolute()

logging.basicConfig()
logging.getLogger().setLevel(logging.DEBUG)
logger = logging.getLogger(__name__)


@retry(None, ConnectionRefusedError)
def _check_tls(ssl_ctx: ssl.SSLContext, port):
    with socket.create_connection(("127.0.0.1", port)) as sock:
        with ssl_ctx.wrap_socket(sock, server_hostname="relukko") as ssock:
            response = ssock.recv(20)
            assert response == b"Hello, TLS Client!"


def _check_has_serial_no(ssl_ctx: ssl.SSLContext):
    serial_numbers = [x["serialNumber"] for x in ssl_ctx.get_ca_certs()]
    assert "4573816C13FE0529EAD797EAB28DFED16930B76B" in serial_numbers


def test_init_relukko_client():
    relukko = RelukkoClient(base_url="", api_key="")

    assert isinstance(relukko, RelukkoClient)
    assert relukko.session.trust_env == True
    assert relukko.session.cookies._cookies == {}
    assert relukko.ssl_ctx is None
    http_adapter: HTTPAdapter = relukko.session.adapters.get("http://")
    assert http_adapter.max_retries.total == 0
    assert http_adapter.max_retries.connect is None
    assert http_adapter.max_retries.read == False
    assert http_adapter.max_retries.redirect is None
    assert http_adapter.max_retries.status is None
    assert http_adapter.max_retries.other is None
    assert http_adapter.max_retries.backoff_factor == 0
    assert http_adapter.max_retries.backoff_max == 120
    assert http_adapter.max_retries.backoff_jitter == 0.0
    assert http_adapter.max_retries.raise_on_redirect == True
    assert http_adapter.max_retries.raise_on_status == True
    assert relukko.tries == 4
    assert relukko.delay == 5
    assert relukko.backoff == 2.0
    assert relukko.max_delay is None
    assert relukko.exceptions == (requests.ConnectionError, RelukkoDoRetry)
    assert relukko.ws_wait_for_timeout == 2
    assert relukko.ws_ping_interval == 60
    assert relukko.ws_ping_timeout == 20
    assert relukko.acquire_wait_for_timeout == 2
    assert relukko.acquire_modulo == 100
    assert relukko.disable_websocket == False
    assert relukko.raise_when_acquire_fails == True

    with pytest.raises(ValueError):
        RelukkoClient(base_url=1000, api_key="")

def test_init_relukko_client_with_env_var(tls_listener):
    _, port = tls_listener

    capath = SCRIPT_DIR / "cert"
    cafile = capath / "rootCA.crt"

    pytest.MonkeyPatch().setenv("REQUESTS_CA_BUNDLE", str(cafile.absolute()))
    relukko = RelukkoClient(base_url="https://relukko", api_key="")
    pytest.MonkeyPatch().delenv("REQUESTS_CA_BUNDLE")
    _check_tls(relukko.ssl_ctx, port)
    _check_has_serial_no(relukko.ssl_ctx)

    pytest.MonkeyPatch().setenv("CURL_CA_BUNDLE", str(cafile.absolute()))
    relukko2 = RelukkoClient(base_url="https://relukko", api_key="")
    pytest.MonkeyPatch().delenv("CURL_CA_BUNDLE")
    _check_tls(relukko2.ssl_ctx, port)
    _check_has_serial_no(relukko2.ssl_ctx)

    pytest.MonkeyPatch().setenv("REQUESTS_CA_BUNDLE", str(capath.absolute()))
    relukko3 = RelukkoClient(base_url="https://relukko", api_key="")
    pytest.MonkeyPatch().delenv("REQUESTS_CA_BUNDLE")
    _check_tls(relukko3.ssl_ctx, port)
    _check_has_serial_no(relukko3.ssl_ctx)

    pytest.MonkeyPatch().setenv("CURL_CA_BUNDLE", str(capath.absolute()))
    relukko4 = RelukkoClient(base_url="https://relukko", api_key="")
    pytest.MonkeyPatch().delenv("CURL_CA_BUNDLE")
    _check_tls(relukko4.ssl_ctx, port)
    _check_has_serial_no(relukko4.ssl_ctx)


def test_init_relukko_client_with_cafile(tls_listener):
    _, port = tls_listener

    cafile = SCRIPT_DIR / "cert" / "rootCA.crt"

    relukko = RelukkoClient(
        base_url="https://relukko", api_key="", cafile=cafile)
    _check_tls(relukko.ssl_ctx, port)
    _check_has_serial_no(relukko.ssl_ctx)


def test_init_relukko_client_with_capath(tls_listener):
    _, port = tls_listener

    capath = SCRIPT_DIR / "cert"

    relukko = RelukkoClient(
        base_url="https://relukko", api_key="", capath=capath)
    _check_tls(relukko.ssl_ctx, port)
    _check_has_serial_no(relukko.ssl_ctx)


def test_init_relukko_client_with_cadata(tls_listener):
    _, port = tls_listener

    cafile = SCRIPT_DIR / "cert" / "rootCA.crt"
    cadata = cafile.read_text()

    relukko = RelukkoClient(
        base_url="https://relukko", api_key="", cadata=cadata)
    _check_tls(relukko.ssl_ctx, port)
    _check_has_serial_no(relukko.ssl_ctx)

    cader = SCRIPT_DIR / "cert" / "rootCA.der"
    cadata = cader.read_bytes()

    relukko2 = RelukkoClient(
        base_url="https://relukko", api_key="", cadata=cadata)
    _check_tls(relukko2.ssl_ctx, port)
    _check_has_serial_no(relukko2.ssl_ctx)


def test_check_tls_throws_error(tls_listener):
    _, port = tls_listener

    relukko = RelukkoClient(
        base_url="https://relukko", api_key="")
    with pytest.raises(ssl.SSLCertVerificationError):
        _check_tls(relukko.ssl_ctx, port)


def test_reconfigure_relukko_client():
    relukko = RelukkoClient(base_url="", api_key="")

    relukko.reconfigure_relukko(
        "https://relukko", api_key="secret-key")

    assert relukko.base_url == parse_url("https://relukko")
    assert relukko.ws_url == parse_url("wss://relukko/ws/broadcast")
    assert relukko.session.headers['X-api-Key'] == "secret-key"

    default_ctx = ssl.create_default_context()
    assert relukko.ssl_ctx.check_hostname == default_ctx.check_hostname
    assert relukko.ssl_ctx.hostname_checks_common_name == \
        default_ctx.hostname_checks_common_name
    assert relukko.ssl_ctx.keylog_filename == default_ctx.keylog_filename
    assert relukko.ssl_ctx.sni_callback == default_ctx.sni_callback
    assert relukko.ssl_ctx.verify_flags == default_ctx.verify_flags
    assert relukko.ssl_ctx.verify_mode == default_ctx.verify_mode
    assert relukko.ssl_ctx.options == default_ctx.options
    assert relukko.tries == 4
    assert relukko.delay == 5
    assert relukko.backoff == 2.0
    assert relukko.max_delay is None
    assert relukko.exceptions == (requests.ConnectionError, RelukkoDoRetry)
    assert relukko.ws_wait_for_timeout == 2
    assert relukko.ws_ping_interval == 60
    assert relukko.ws_ping_timeout == 20
    assert relukko.acquire_wait_for_timeout == 2
    assert relukko.acquire_modulo == 100
    assert relukko.disable_websocket == False
    assert relukko.raise_when_acquire_fails == True


def test_reconfigure_relukko_client_extended():
    relukko = RelukkoClient(base_url="", api_key="")

    cookies = cookiejar_from_dict({"cookie1": "value1"})
    verify_mode = ssl.VerifyMode.CERT_NONE
    verify_flags = ssl.VerifyFlags.VERIFY_DEFAULT
    options = ssl.Options.OP_ALL
    ws_wait_for_timeout = 27
    ws_ping_interval = 600
    ws_ping_timeout = 23
    acquire_wait_for_timeout = 22
    acquire_modulo = 202

    relukko.reconfigure_relukko(
        base_url="https://relukko", api_key="my-API-key", trust_env=False,
        cookies=cookies, total=100, connect=99, read=98, redirect=97,
        status=96, other=95, backoff_factor=94, backoff_max=1000,
        backoff_jitter=93, raise_on_redirect=False, raise_on_status=False,
        check_hostname=False, hostname_checks_common_name=False,
        verify_mode=verify_mode, verify_flags=verify_flags, options=options,
        ws_wait_for_timeout=ws_wait_for_timeout, acquire_modulo=acquire_modulo,
        ws_ping_interval=ws_ping_interval, ws_ping_timeout=ws_ping_timeout,
        acquire_wait_for_timeout=acquire_wait_for_timeout,
        disable_websocket=True, raise_when_acquire_fails=False,
    )

    assert relukko.session.trust_env == False
    assert relukko.session.cookies == cookies
    assert relukko.base_url == parse_url("https://relukko")
    assert relukko.ws_url == parse_url("wss://relukko/ws/broadcast")
    assert relukko.session.headers['X-api-Key'] == "my-API-key"
    http_adapter: HTTPAdapter = relukko.session.adapters.get("http://")
    assert http_adapter.max_retries.total == 100
    assert http_adapter.max_retries.connect == 99
    assert http_adapter.max_retries.read == 98
    assert http_adapter.max_retries.redirect == 97
    assert http_adapter.max_retries.status == 96
    assert http_adapter.max_retries.other == 95
    assert http_adapter.max_retries.backoff_factor == 94.0
    assert http_adapter.max_retries.backoff_max == 1000.0
    assert http_adapter.max_retries.backoff_jitter == 93.0
    assert http_adapter.max_retries.raise_on_redirect == False
    assert http_adapter.max_retries.raise_on_status == False
    assert isinstance(relukko.ssl_ctx, ssl.SSLContext)
    assert relukko.ssl_ctx.check_hostname == False
    assert relukko.ssl_ctx.hostname_checks_common_name == False
    assert relukko.ssl_ctx.keylog_filename is None
    assert relukko.ssl_ctx.sni_callback is None
    assert relukko.ssl_ctx.verify_flags == verify_flags
    assert relukko.ssl_ctx.verify_mode == verify_mode
    assert relukko.ssl_ctx.options == options
    assert relukko.tries == 4
    assert relukko.delay == 5
    assert relukko.backoff == 2.0
    assert relukko.max_delay is None
    assert relukko.exceptions == (requests.ConnectionError, RelukkoDoRetry)
    assert relukko.ws_wait_for_timeout == ws_wait_for_timeout
    assert relukko.ws_ping_interval == ws_ping_interval
    assert relukko.ws_ping_timeout == ws_ping_timeout
    assert relukko.acquire_wait_for_timeout == acquire_wait_for_timeout
    assert relukko.acquire_modulo == acquire_modulo
    assert relukko.disable_websocket == True
    assert relukko.raise_when_acquire_fails == False


def test_init_relukko_client_extented():
    cookies = cookiejar_from_dict({"cookie1": "value1"})
    ws_wait_for_timeout = 27
    ws_ping_interval = 600
    ws_ping_timeout = 23
    acquire_wait_for_timeout = 22
    acquire_modulo = 202

    relukko = RelukkoClient(
        base_url="http://relukko", api_key="my-API-key", trust_env=False,
        cookies=cookies, total=100, connect=99, read=98, redirect=97,
        status=96, other=95, backoff_factor=94, backoff_max=1000,
        backoff_jitter=93, raise_on_redirect=False, raise_on_status=False,
        ws_wait_for_timeout=ws_wait_for_timeout, acquire_modulo=acquire_modulo,
        ws_ping_interval=ws_ping_interval, ws_ping_timeout=ws_ping_timeout,
        acquire_wait_for_timeout=acquire_wait_for_timeout,
        disable_websocket=True, raise_when_acquire_fails=False,
    )

    assert relukko.session.trust_env == False
    assert relukko.session.cookies == cookies
    assert relukko.base_url == parse_url("http://relukko")
    assert relukko.ws_url == parse_url("ws://relukko/ws/broadcast")
    assert relukko.session.headers['X-api-Key'] == "my-API-key"
    assert relukko.ssl_ctx is None
    http_adapter: HTTPAdapter = relukko.session.adapters.get("http://")
    assert http_adapter.max_retries.total == 100
    assert http_adapter.max_retries.connect == 99
    assert http_adapter.max_retries.read == 98
    assert http_adapter.max_retries.redirect == 97
    assert http_adapter.max_retries.status == 96
    assert http_adapter.max_retries.other == 95
    assert http_adapter.max_retries.backoff_factor == 94.0
    assert http_adapter.max_retries.backoff_max == 1000.0
    assert http_adapter.max_retries.backoff_jitter == 93.0
    assert http_adapter.max_retries.raise_on_redirect == False
    assert http_adapter.max_retries.raise_on_status == False
    assert relukko.tries == 4
    assert relukko.delay == 5
    assert relukko.backoff == 2.0
    assert relukko.max_delay is None
    assert relukko.exceptions == (requests.ConnectionError, RelukkoDoRetry)
    assert relukko.ws_wait_for_timeout == ws_wait_for_timeout
    assert relukko.ws_ping_interval == ws_ping_interval
    assert relukko.ws_ping_timeout == ws_ping_timeout
    assert relukko.acquire_wait_for_timeout == acquire_wait_for_timeout
    assert relukko.acquire_modulo == acquire_modulo
    assert relukko.disable_websocket == True
    assert relukko.raise_when_acquire_fails == False


def test_init_relukko_client_ssl_ctx():
    relukko = RelukkoClient(
        base_url="https://relukko", api_key="my-API-key",
    )

    default_ctx = ssl.create_default_context()

    assert isinstance(relukko.ssl_ctx, ssl.SSLContext)
    assert relukko.ssl_ctx.check_hostname == default_ctx.check_hostname
    assert relukko.ssl_ctx.hostname_checks_common_name == \
        default_ctx.hostname_checks_common_name
    assert relukko.ssl_ctx.keylog_filename == default_ctx.keylog_filename
    assert relukko.ssl_ctx.sni_callback == default_ctx.sni_callback
    assert relukko.ssl_ctx.verify_flags == default_ctx.verify_flags
    assert relukko.ssl_ctx.verify_mode == default_ctx.verify_mode
    assert relukko.ssl_ctx.options == default_ctx.options
    assert relukko.tries == 4
    assert relukko.delay == 5
    assert relukko.backoff == 2.0
    assert relukko.max_delay is None
    assert relukko.exceptions == (requests.ConnectionError, RelukkoDoRetry)


def test_init_relukko_client_ssl_ctx_extended():
    verify_mode = ssl.VerifyMode.CERT_NONE
    verify_flags = ssl.VerifyFlags.VERIFY_DEFAULT
    options = ssl.Options.OP_ALL

    relukko = RelukkoClient(
        base_url="https://relukko", api_key="my-API-key", check_hostname=False,
        hostname_checks_common_name=False, verify_mode=verify_mode,
        verify_flags=verify_flags, options=options
    )
    assert isinstance(relukko.ssl_ctx, ssl.SSLContext)
    assert relukko.ssl_ctx.check_hostname == False
    assert relukko.ssl_ctx.hostname_checks_common_name == False
    assert relukko.ssl_ctx.keylog_filename is None
    assert relukko.ssl_ctx.sni_callback is None
    assert relukko.ssl_ctx.verify_flags == verify_flags
    assert relukko.ssl_ctx.verify_mode == verify_mode
    assert relukko.ssl_ctx.options == options
    assert relukko.tries == 4
    assert relukko.delay == 5
    assert relukko.backoff == 2.0
    assert relukko.max_delay is None
    assert relukko.exceptions == (requests.ConnectionError, RelukkoDoRetry)


def test_init_relukko_client_retry():

    exceptions = (
        ValueError,
        ConnectionRefusedError,
        ConnectionResetError,
    )

    relukko = RelukkoClient(
        base_url="", api_key="", tries=100, delay=99, backoff=98.7,
        max_delay=97, exceptions=exceptions)

    assert relukko.tries == 100
    assert relukko.delay == 99
    assert relukko.backoff == 98.7
    assert relukko.max_delay == 97
    assert relukko.exceptions == exceptions


def test_reconfigre_relukko_client_retry():
    relukko = RelukkoClient(base_url="", api_key="")

    exceptions = (
        ValueError,
        ConnectionRefusedError,
        ConnectionResetError,
    )

    relukko.reconfigure_relukko(
        tries=100, delay=99, backoff=98.7,
        max_delay=97, exceptions=exceptions)

    assert relukko.tries == 100
    assert relukko.delay == 99
    assert relukko.backoff == 98.7
    assert relukko.max_delay == 97
    assert relukko.exceptions == exceptions


def test_aqcuire_http_status_code_400_with_raise(requests_mock):
    requests_mock.register_uri(
        ANY,
        ANY,
        status_code=400,
        reason="Bad Request",
        json={"status": 400, "message": "Bad Request"},
    )

    relukko = RelukkoClient(
        base_url="http://relukko", api_key="key",
        delay=1, backoff=1.01, disable_websocket=True,
    )
    with pytest.raises(requests.HTTPError):
        relukko.acquire_relukko("400Lock", "PyTest", 300)


def test_aqcuire_http_status_code_400_without_raise(requests_mock):
    requests_mock.register_uri(
        ANY,
        ANY,
        status_code=400,
        reason="Bad Request",
        json={"status": 400, "message": "Bad Request"},
    )

    relukko = RelukkoClient(
        base_url="http://relukko", api_key="key",disable_websocket=True,
        delay=1, backoff=1.01, raise_when_acquire_fails=False,
    )
    lock = relukko.acquire_relukko("500Lock", "PyTest", 300)
    assert isinstance(lock, requests.exceptions.HTTPError)


def test_aqcuire_http_status_code_418_with_raise(requests_mock):
    requests_mock.register_uri(
        ANY,
        ANY,
        status_code=418,
        reason="I'm a teapot",
        text="The server refuses the attempt to brew coffee with a teapot.",
    )

    relukko = RelukkoClient(
        base_url="http://relukko", api_key="key",
        delay=1, backoff=1.01, disable_websocket=True,
    )
    with pytest.raises(RuntimeError):
        relukko.acquire_relukko("418Lock", "PyTest", 300)


def test_aqcuire_http_status_code_418_without_raise(requests_mock):
    requests_mock.register_uri(
        ANY,
        ANY,
        status_code=418,
        reason="I'm a teapot",
        text="The server refuses the attempt to brew coffee with a teapot.",
    )

    relukko = RelukkoClient(
        base_url="http://relukko", api_key="key",disable_websocket=True,
        delay=1, backoff=1.01, raise_when_acquire_fails=False,
    )
    lock = relukko.acquire_relukko("418Lock", "PyTest", 300)
    assert isinstance(lock, RuntimeError)


def test_aqcuire_http_status_code_500_with_raise(requests_mock):
    requests_mock.register_uri(
        ANY,
        ANY,
        status_code=500,
        reason="Internal Server Error",
    )

    relukko = RelukkoClient(
        base_url="http://relukko", api_key="key",
        delay=1, backoff=1.01, disable_websocket=True,
    )
    with pytest.raises(RelukkoDoRetry):
        relukko.acquire_relukko("500Lock", "PyTest", 300)


def test_aqcuire_http_status_code_500_without_raise(requests_mock):
    requests_mock.register_uri(
        ANY,
        ANY,
        status_code=500,
        reason="Internal Server Error",
    )

    relukko = RelukkoClient(
        base_url="http://relukko", api_key="key",disable_websocket=True,
        delay=1, backoff=1.01, raise_when_acquire_fails=False,
    )
    lock = relukko.acquire_relukko("500Lock", "PyTest", 300)
    assert isinstance(lock, RelukkoDoRetry)


def test_aqcuire_relukko_retry_with_max_delay(requests_mock, caplog):
    requests_mock.register_uri(
        ANY,
        ANY,
        status_code=500,
        reason="Internal Server Error",
    )

    relukko = RelukkoClient(
        base_url="http://relukko", api_key="key",disable_websocket=True,
        delay=1, backoff=3, raise_when_acquire_fails=False,
        max_delay=2,
    )

    lock = relukko.acquire_relukko("MaxDelayLock", "PyTest", 300)

    one_sec_count = 0
    two_sec_count = 0
    for record in caplog.records:
        if "Retrying in 2 seconds..." in record.message:
            two_sec_count += 1
        elif "Retrying in 1 seconds..." in record.message:
            one_sec_count += 1

    assert one_sec_count == 1
    assert two_sec_count == 2
    assert isinstance(lock, RelukkoDoRetry)


def test_get_relukko_retry_400(requests_mock):
    requests_mock.register_uri(
        ANY,
        ANY,
        status_code=400,
        reason="Bad Request",
        json={"status": 400, "message": "Bad Request"},
    )

    relukko = RelukkoClient(
        base_url="http://relukko", api_key="key",disable_websocket=True,
        delay=1, backoff=0.5, raise_when_acquire_fails=False,
    )

    with pytest.raises(requests.HTTPError):
        relukko.get_lock("GetLock400")


def test_get_relukko_retry_418(requests_mock):
    requests_mock.register_uri(
        ANY,
        ANY,
        status_code=418,
        reason="I'm a teapot",
        text="The server refuses the attempt to brew coffee with a teapot.",
    )

    relukko = RelukkoClient(
        base_url="http://relukko", api_key="key",disable_websocket=True,
        delay=1, backoff=0.5, raise_when_acquire_fails=False,
    )

    with pytest.raises(RuntimeError):
        relukko.get_lock("GetLock418")


def test_get_relukko_retry_500(requests_mock):
    requests_mock.register_uri(
        ANY,
        ANY,
        status_code=500,
        reason="Internal Server Error",
    )

    relukko = RelukkoClient(
        base_url="http://relukko", api_key="key",disable_websocket=True,
        delay=1, backoff=0.5, raise_when_acquire_fails=False,
    )

    with pytest.raises(RelukkoDoRetry):
        relukko.get_lock("GetLock500")


def test_acquire_relukko(relukko_backend):
    relukko, _ = relukko_backend

    lock = relukko.acquire_relukko("pylock", "pytest", 10)
    assert isinstance(lock, Dict)
    start_time = time.time()
    lock = relukko.acquire_relukko("pylock", "pytest", 30)
    end_time = time.time()
    assert lock is None
    assert 29 < end_time - start_time < 37


def test_acquire_relukko_with_del(relukko_backend):
    relukko, _ = relukko_backend
    lock = relukko.acquire_relukko("Pylock", "pytest", 10)
    id = lock['id']

    expires_at = datetime.now(timezone.utc) - timedelta(minutes=2, seconds=45)
    upd_lock = relukko.update_relukko(id, expires_at=expires_at)
    assert id == upd_lock['id']

    start_time = time.time()
    lock = relukko.acquire_relukko("Pylock", "pytest", 60)
    end_time = time.time()

    assert lock is not None
    assert upd_lock['id'] != lock['id']
    assert 14 < end_time - start_time < 50


def test_delete_relukko(relukko_backend):
    relukko, _ = relukko_backend

    lock = relukko.acquire_relukko("del_lock", "pytest", 10)
    del_lock = relukko.delete_relukko(lock['id'])
    assert lock == del_lock

    get_lock = relukko.get_lock(lock['id'])
    assert get_lock['status'] == 422

    locks = relukko.get_locks()
    assert lock not in locks


def test_get_relukko(relukko_backend):
    relukko, _ = relukko_backend

    lock = relukko.acquire_relukko("get_lock", "pytest", 10)
    assert isinstance(lock, Dict)

    get_lock = relukko.get_lock(lock['id'])
    assert get_lock == lock

    locks = relukko.get_locks()
    assert lock in locks


def test_update_relukko_creator(relukko_backend):
    relukko, _ = relukko_backend

    lock = relukko.acquire_relukko("update_lock", "pytest", 10)
    assert lock['creator'] == "pytest"

    upd_lock = relukko.update_relukko(lock['id'], creator="tsetyp")
    assert upd_lock['creator'] == "tsetyp"

    get_lock = relukko.get_lock(lock['id'])
    assert get_lock['creator'] == "tsetyp"


def test_update_relukko_expires_at(relukko_backend):
    relukko, _ = relukko_backend

    lock = relukko.acquire_relukko("update_lock_exp", "pytest", 10)
    id = lock['id']

    expires_at = datetime.fromisoformat("2099-12-31T12:34:56.789Z")
    upd_lock = relukko.update_relukko(id, expires_at=expires_at)
    assert upd_lock['expires_at'] == "2099-12-31T12:34:56.789Z"

    with pytest.raises(ValueError):
        relukko.update_relukko(id, expires_at="2099-12-01T23:12Z")


def test_keep_relukko_alive(relukko_backend):
    relukko, _ = relukko_backend

    lock = relukko.acquire_relukko("keep_me", "pytest", 10)
    id = lock['id']

    start_time = datetime.now(timezone.utc)
    keep_lock = relukko.keep_relukko_alive(id)
    assert lock['expires_at'] is not keep_lock['expires_at']

    expires_at = datetime.fromisoformat(keep_lock['expires_at'])

    expires_diff = expires_at - start_time
    assert 295 < expires_diff.seconds < 305


def test_keep_relukko_alive_put(relukko_backend):
    relukko, _ = relukko_backend

    lock = relukko.acquire_relukko("keep_me_more", "pytest", 10)
    id = lock['id']

    start_time = datetime.now(timezone.utc)
    keep_lock = relukko.keep_relukko_alive_put(id, 3600)
    assert lock['expires_at'] is not keep_lock['expires_at']

    expires_at = datetime.fromisoformat(keep_lock['expires_at'])

    expires_diff = expires_at - start_time
    assert 3595 < expires_diff.seconds < 3605


def test_add_to_expires_at_time(relukko_backend):
    relukko, _ = relukko_backend
    lock = relukko.acquire_relukko("add_me_up", "pytest", 10)
    id = lock['id']
    start_expires_at = datetime.fromisoformat(lock['expires_at'])

    add_lock = relukko.add_to_expires_at_time(id)
    assert lock['expires_at'] is not add_lock['expires_at']

    add_expires_at = datetime.fromisoformat(add_lock['expires_at'])

    expires_diff = add_expires_at - start_expires_at
    assert expires_diff.seconds == 300


def test_add_to_expires_at_time_put(relukko_backend):
    relukko, _ = relukko_backend
    lock = relukko.acquire_relukko("add_me_up_more", "pytest", 10)
    id = lock['id']
    start_expires_at = datetime.fromisoformat(lock['expires_at'])

    add_lock = relukko.add_to_expires_at_time_put(id, 3600)
    assert lock['expires_at'] is not add_lock['expires_at']

    add_expires_at = datetime.fromisoformat(add_lock['expires_at'])

    expires_diff = add_expires_at - start_expires_at
    assert expires_diff.seconds == 3600
